@@ -1,6 +1,6 @@
 /* copypass.c - cpio copy pass sub-function.
-   Copyright (C) 1990, 1991, 1992, 2001, 2003, 2004, 2006, 2007, 2010
-   Free Software Foundation, Inc.
+   Copyright (C) 1990-1992, 2001, 2003-2004, 2006-2007, 2010, 2014-2015,
+   2017 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -60,10 +60,6 @@ process_copy_pass ()
   int in_file_des;		/* Input file descriptor.  */
   int out_file_des;		/* Output file descriptor.  */
   int existing_dir;		/* True if file is a dir & already exists.  */
-#ifdef HPUX_CDF
-  int cdf_flag;
-  int cdf_char;
-#endif
 
   newdir_umask = umask (0);     /* Reset umask to preserve modes of
 				   created files  */
@@ -120,13 +116,6 @@ process_copy_pass ()
       /* Make the name of the new file.  */
       for (slash = input_name.ds_string; *slash == '/'; ++slash)
 	;
-#ifdef HPUX_CDF
-      /* For CDF's we add a 2nd `/' after all "hidden" directories.
-	 This kind of a kludge, but it's what we do when creating
-	 archives, and it's easier to do this than to separately
-	 keep track of which directories in a path are "hidden".  */
-      slash = add_cdf_double_slashes (slash);
-#endif
       ds_resize (&output_name, dirname_len + strlen (slash) + 2);
       strcpy (output_name.ds_string + dirname_len + 1, slash);
 
@@ -200,17 +189,8 @@ process_copy_pass ()
 		}
 
 	      copy_files_disk_to_disk (in_file_des, out_file_des, in_file_stat.st_size, input_name.ds_string);
-	      disk_empty_output_buffer (out_file_des);
-	      /* Debian hack to fix a bug in the --sparse option.
-                 This bug has been reported to
-                 "bug-gnu-utils@prep.ai.mit.edu".  (96/7/10) -BEM */
-	      if (delayed_seek_count > 0)
-		{
-		  lseek (out_file_des, delayed_seek_count-1, SEEK_CUR);
-		  write (out_file_des, "", 1);
-		  delayed_seek_count = 0;
-		}
-
+	      disk_empty_output_buffer (out_file_des, true);
+	      
 	      set_copypass_perms (out_file_des,
 				  output_name.ds_string, &in_file_stat);
 
@@ -392,7 +372,7 @@ link_to_maj_min_ino (char *file_name, int st_dev_maj, int st_dev_min,
    is created, -1 otherwise.  */
 
 int
-link_to_name (char *link_name, char *link_target)
+link_to_name (char const *link_name, char const *link_target)
 {
   int res = link (link_target, link_name);
   if (res < 0 && create_dir_flag)
