@@ -1,5 +1,5 @@
 /* Extended cpio header from POSIX.1.
-   Copyright (C) 1992, 2006-2007, 2010, 2014-2015 Free Software
+   Copyright (C) 1992, 2006-2007, 2010, 2014-2015, 2017 Free Software
    Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
@@ -126,8 +126,15 @@ struct cpio_file_stat /* Internal representation of a CPIO header */
   size_t c_namesize;
   uint32_t c_chksum;
   char *c_name;
-  char *c_tar_linkname;
+  size_t c_name_buflen;
+  char const *c_tar_linkname;
 };
 
+#define CPIO_FILE_STAT_INITIALIZER \
+  { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, 0, NULL }
+void cpio_file_stat_init (struct cpio_file_stat *file_hdr);
+void cpio_file_stat_free (struct cpio_file_stat *file_hdr);
+void cpio_set_c_name(struct cpio_file_stat *file_hdr, char *name);
+void cpio_realloc_c_name (struct cpio_file_stat *file_hdr, size_t len);
 
 #endif /* cpiohdr.h */
