@@ -1,6 +1,6 @@
 /* This file is part of GNU cpio.
-   Copyright (C) 2005, 2007, 2010, 2014-2015 Free Software Foundation,
-   Inc.
+   Copyright (C) 2005, 2007, 2010, 2014-2015, 2017 Free Software
+   Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
