@@ -1,6 +1,6 @@
 /* userspec.c -- Parse a user and group string.
-   Copyright (C) 1989, 1990, 1991, 1992, 2001, 
-   2004, 2005, 2007 Free Software Foundation, Inc.
+   Copyright (C) 1989-1992, 2001, 2004-2005, 2007, 2010, 2014-2015, 2017
+   Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -20,36 +20,10 @@
 /* Written by David MacKenzie <djm@gnu.ai.mit.edu>.  */
 
 #include <system.h>
-
-#ifdef __GNUC__
-#define alloca __builtin_alloca
-#else
-#ifdef HAVE_ALLOCA_H
 #include <alloca.h>
-#else
-#ifdef _AIX
- #pragma alloca
-#else
-char *alloca ();
-#endif
-#endif
-#endif
-
 #include <stdio.h>
 #include <ctype.h>
 #include <sys/types.h>
-#include <pwd.h>
-#include <grp.h>
-
-#if !HAVE_DECL_GETPWNAM
-extern struct passwd *getpwnam (const char *name);
-#endif
-#if !HAVE_DECL_GETGRNAM
-extern struct group *getgrnam (const char *name);
-#endif
-#if !HAVE_DECL_GETGRGID
-extern struct group *getgrgid (gid_t gid);
-#endif
 
 #ifndef HAVE_ENDPWENT
 # define endpwent()
@@ -134,10 +108,16 @@ parse_user_spec (const char *spec_arg, uid_t *uid, gid_t *gid,
 
   if (u != NULL)
     {
-      pwd = getpwnam (u);
+      if (*u == '+')
+	{
+	  pwd = NULL;
+	  ++u;
+	}
+      else
+	pwd = getpwnam (u);
+
       if (pwd == NULL)
 	{
-
 	  if (!isnumber_p (u))
 	    error_msg = _("invalid user");
 	  else
@@ -181,7 +161,14 @@ parse_user_spec (const char *spec_arg, uid_t *uid, gid_t *gid,
   if (g != NULL && error_msg == NULL)
     {
       /* Explicit group.  */
-      grp = getgrnam (g);
+      if (*g == '+')
+	{
+	  grp = NULL;
+	  ++g;
+	}
+      else
+	grp = getgrnam (g);
+
       if (grp == NULL)
 	{
 	  if (!isnumber_p (g))
