@@ -1,6 +1,6 @@
 /* util.c - Several utility routines for cpio.
-   Copyright (C) 1990, 1991, 1992, 2001, 2004,
-   2006, 2007 Free Software Foundation, Inc.
+   Copyright (C) 1990-1992, 2001, 2004, 2006-2007, 2010-2011, 2014-2015,
+   2017 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -90,17 +90,17 @@ tape_empty_output_buffer (int out_des)
 	  rest_bytes_written = rmtwrite (out_des, output_buffer,
 					 rest_output_size);
 	  if (rest_bytes_written != rest_output_size)
-	    error (1, errno, _("write error"));
+	    error (PAXEXIT_FAILURE, errno, _("write error"));
 	}
       else
-	error (1, errno, _("write error"));
+	error (PAXEXIT_FAILURE, errno, _("write error"));
     }
   output_bytes += output_size;
   out_buff = output_buffer;
   output_size = 0;
 }
 
-static int sparse_write (int fildes, char *buf, unsigned int nbyte);
+static ssize_t sparse_write (int fildes, char *buf, size_t nbyte, bool flush);
 
 /* Write `output_size' bytes of `output_buffer' to file
    descriptor OUT_DES and reset `output_size' and `out_buff'.
@@ -113,9 +113,9 @@ static int sparse_write (int fildes, char *buf, unsigned int nbyte);
    insure this.  */
 
 void
-disk_empty_output_buffer (int out_des)
+disk_empty_output_buffer (int out_des, bool flush)
 {
-  int bytes_written;
+  ssize_t bytes_written;
 
   if (swapping_halfwords || swapping_bytes)
     {
@@ -136,13 +136,16 @@ disk_empty_output_buffer (int out_des)
     }
 
   if (sparse_flag)
-    bytes_written = sparse_write (out_des, output_buffer, output_size);
+    bytes_written = sparse_write (out_des, output_buffer, output_size, flush);
   else
     bytes_written = write (out_des, output_buffer, output_size);
 
   if (bytes_written != output_size)
     {
-      error (1, errno, _("write error"));
+      if (bytes_written == -1)
+	error (PAXEXIT_FAILURE, errno, _("write error"));
+      else
+	error (PAXEXIT_FAILURE, 0, _("write error: partial write"));
     }
   output_bytes += output_size;
   out_buff = output_buffer;
@@ -203,13 +206,10 @@ tape_fill_input_buffer (int in_des, int num_bytes)
       get_next_reel (in_des);
       input_size = rmtread (in_des, input_buffer, num_bytes);
     }
-  if (input_size < 0)
-    error (1, errno, _("read error"));
+  if (input_size == SAFE_READ_ERROR)
+    error (PAXEXIT_FAILURE, errno, _("read error"));
   if (input_size == 0)
-    {
-      error (0, 0, _("premature end of file"));
-      exit (1);
-    }
+    error (PAXEXIT_FAILURE, 0, _("premature end of file"));
   input_bytes += input_size;
 }
 
@@ -224,7 +224,7 @@ disk_fill_input_buffer (int in_des, off_t num_bytes)
   in_buff = input_buffer;
   num_bytes = (num_bytes < DISK_IO_BLOCK_SIZE) ? num_bytes : DISK_IO_BLOCK_SIZE;
   input_size = read (in_des, input_buffer, num_bytes);
-  if (input_size < 0) 
+  if (input_size == SAFE_READ_ERROR)
     {
       input_size = 0;
       return (-1);
@@ -275,7 +275,7 @@ disk_buffered_write (char *in_buf, int out_des, off_t num_bytes)
     {
       space_left = DISK_IO_BLOCK_SIZE - output_size;
       if (space_left == 0)
-	disk_empty_output_buffer (out_des);
+	disk_empty_output_buffer (out_des, false);
       else
 	{
 	  if (bytes_left < space_left)
@@ -373,7 +373,7 @@ tape_buffered_peek (char *peek_buf, int in_des, int num_bytes)
 	    break;
 	}
       if (tmp_input_size < 0)
-	error (1, errno, _("read error"));
+	error (PAXEXIT_FAILURE, errno, _("read error"));
       input_bytes += tmp_input_size;
       input_size += tmp_input_size;
     }
@@ -484,9 +484,9 @@ copy_files_disk_to_tape (int in_des, int out_des, off_t num_bytes,
   while (num_bytes > 0)
     {
       if (input_size == 0)
-	if (rc = disk_fill_input_buffer (in_des,
-					 num_bytes < DISK_IO_BLOCK_SIZE ?
-					 num_bytes : DISK_IO_BLOCK_SIZE))
+	if ((rc = disk_fill_input_buffer (in_des,
+					  num_bytes < DISK_IO_BLOCK_SIZE ?
+					  num_bytes : DISK_IO_BLOCK_SIZE)))
 	  {
 	    if (rc > 0)
 	      {
@@ -498,8 +498,9 @@ copy_files_disk_to_tape (int in_des, int out_des, off_t num_bytes,
 			 filename,  STRINGIFY_BIGINT (num_bytes, buf));
 	      }
 	    else
-	      error (0, 0, _("Read error at byte %lld in file %s, padding with zeros"),
-			original_num_bytes - num_bytes, filename);
+	      error (0, 0,
+		     _("Read error at byte %lld in file %s, padding with zeros"),
+		     (long long) (original_num_bytes - num_bytes), filename);
 	    write_nuls_to_file (num_bytes, out_des, tape_buffered_write);
 	    break;
 	  }
@@ -527,8 +528,8 @@ void
 copy_files_disk_to_disk (int in_des, int out_des, off_t num_bytes,
 			 char *filename)
 {
-  long size;
-  long k;
+  off_t size;
+  off_t k;
   off_t original_num_bytes;
   int rc;
 
@@ -536,7 +537,7 @@ copy_files_disk_to_disk (int in_des, int out_des, off_t num_bytes,
   while (num_bytes > 0)
     {
       if (input_size == 0)
-	if (rc = disk_fill_input_buffer (in_des, num_bytes))
+	if ((rc = disk_fill_input_buffer (in_des, num_bytes)))
 	  {
 	    if (rc > 0)
 	      {
@@ -548,8 +549,9 @@ copy_files_disk_to_disk (int in_des, int out_des, off_t num_bytes,
 		       filename,  STRINGIFY_BIGINT (num_bytes, buf));
 	      }
 	    else
-	      error (0, 0, _("Read error at byte %lld in file %s, padding with zeros"),
-			original_num_bytes - num_bytes, filename);
+	      error (0, 0,
+		     _("Read error at byte %lld in file %s, padding with zeros"),
+		     (long long) (original_num_bytes - num_bytes), filename);
 	    write_nuls_to_file (num_bytes, out_des, disk_buffered_write);
 	    break;
 	  }
@@ -569,7 +571,7 @@ copy_files_disk_to_disk (int in_des, int out_des, off_t num_bytes,
 /* Warn if file changed while it was being copied.  */
 
 void
-warn_if_file_changed (char *file_name, unsigned long old_file_size,
+warn_if_file_changed (char *file_name, off_t old_file_size,
 		      time_t old_file_mtime)
 {
   struct stat new_file_stat;
@@ -596,28 +598,14 @@ warn_if_file_changed (char *file_name, unsigned long old_file_size,
    Do not destroy any nondirectories while creating directories.  */
 
 void
-create_all_directories (char *name)
+create_all_directories (char const *name)
 {
   char *dir;
-  int   mode;
-#ifdef HPUX_CDF
-  int   cdf;
-#endif
 
   dir = dir_name (name);
-  mode = 0700;
-#ifdef HPUX_CDF
-  cdf = islastparentcdf (name);
-  if (cdf)
-    {
-      dir [strlen (dir) - 1] = '\0';	/* remove final + */
-      mode = 04700;
-    }
-  
-#endif
   
   if (dir == NULL)
-    error (2, 0, _("virtual memory exhausted"));
+    error (PAXEXIT_FAILURE, 0, _("virtual memory exhausted"));
 
   if (dir[0] != '.' || dir[1] != '\0')
     {
@@ -660,13 +648,13 @@ prepare_append (int out_file_des)
   start_of_block = start_of_header - useful_bytes_in_block;
 
   if (lseek (out_file_des, start_of_block, SEEK_SET) < 0)
-    error (1, errno, _("cannot seek on output"));
+    error (PAXEXIT_FAILURE, errno, _("cannot seek on output"));
   if (useful_bytes_in_block > 0)
     {
       tmp_buf = (char *) xmalloc (useful_bytes_in_block);
       read (out_file_des, tmp_buf, useful_bytes_in_block);
       if (lseek (out_file_des, start_of_block, SEEK_SET) < 0)
-	error (1, errno, _("cannot seek on output"));
+	error (PAXEXIT_FAILURE, errno, _("cannot seek on output"));
       /* fix juo -- is this copy_tape_buf_out?  or copy_disk? */
       tape_buffered_write (tmp_buf, out_file_des, useful_bytes_in_block);
       free (tmp_buf);
@@ -686,9 +674,10 @@ prepare_append (int out_file_des)
 
 struct inode_val
 {
-  unsigned long inode;
+  ino_t inode;
   unsigned long major_num;
   unsigned long minor_num;
+  ino_t trans_inode;
   char *file_name;
 };
 
@@ -712,12 +701,11 @@ inode_val_compare (const void *val1, const void *val2)
          && ival1->minor_num == ival2->minor_num;
 }
 
-char *
-find_inode_file (unsigned long node_num, unsigned long major_num,
+static struct inode_val *
+find_inode_val (ino_t node_num, unsigned long major_num,
 		 unsigned long minor_num)
 {
   struct inode_val sample;
-  struct inode_val *ival;
   
   if (!hash_table)
     return NULL;
@@ -725,32 +713,78 @@ find_inode_file (unsigned long node_num, unsigned long major_num,
   sample.inode = node_num;
   sample.major_num = major_num;
   sample.minor_num = minor_num;
-  ival = hash_lookup (hash_table, &sample);
+  return hash_lookup (hash_table, &sample);
+}
+
+char *
+find_inode_file (ino_t node_num, unsigned long major_num,
+		 unsigned long minor_num)
+{
+  struct inode_val *ival = find_inode_val (node_num, major_num, minor_num);
   return ival ? ival->file_name : NULL;
 }
 
 /* Associate FILE_NAME with the inode NODE_NUM.  (Insert into hash table.)  */
 
-void
-add_inode (unsigned long node_num, char *file_name, unsigned long major_num,
+static ino_t next_inode;
+
+struct inode_val *
+add_inode (ino_t node_num, char *file_name, unsigned long major_num,
 	   unsigned long minor_num)
 {
   struct inode_val *temp;
-  struct inode_val *e;
+  struct inode_val *e = NULL;
   
   /* Create new inode record.  */
   temp = (struct inode_val *) xmalloc (sizeof (struct inode_val));
   temp->inode = node_num;
   temp->major_num = major_num;
   temp->minor_num = minor_num;
-  temp->file_name = xstrdup (file_name);
+  temp->file_name = file_name ? xstrdup (file_name) : NULL;
+
+  if (renumber_inodes_option)
+    temp->trans_inode = next_inode++;
+  else
+    temp->trans_inode = temp->inode;
 
   if (!((hash_table
 	 || (hash_table = hash_initialize (0, 0, inode_val_hasher,
 					   inode_val_compare, 0)))
 	&& (e = hash_insert (hash_table, temp))))
     xalloc_die ();
-  /* FIXME: e is not used */
+  return e;
+}
+
+static void
+get_inode_and_dev (struct cpio_file_stat *hdr, struct stat *st)
+{
+  if (renumber_inodes_option)
+    {
+      if (st->st_nlink > 1)
+	{
+	  struct inode_val *ival = find_inode_val (st->st_ino,
+						   major (st->st_dev),
+						   minor (st->st_dev));
+	  if (!ival)
+	    ival = add_inode (st->st_ino, NULL,
+			      major (st->st_dev), minor (st->st_dev));
+	  hdr->c_ino = ival->trans_inode;
+	}
+      else
+	hdr->c_ino = next_inode++;
+    }
+  else
+    hdr->c_ino = st->st_ino;
+  if (ignore_devno_option)
+    {
+      hdr->c_dev_maj = 0;
+      hdr->c_dev_min = 0;
+    }
+  else
+    {
+      hdr->c_dev_maj = major (st->st_dev);
+      hdr->c_dev_min = minor (st->st_dev);
+    }
 }
 
 
@@ -820,10 +854,10 @@ get_next_reel (int tape_des)
   /* Open files for interactive communication.  */
   tty_in = fopen (TTY_NAME, "r");
   if (tty_in == NULL)
-    error (2, errno, TTY_NAME);
+    error (PAXEXIT_FAILURE, errno, TTY_NAME);
   tty_out = fopen (TTY_NAME, "w");
   if (tty_out == NULL)
-    error (2, errno, TTY_NAME);
+    error (PAXEXIT_FAILURE, errno, TTY_NAME);
 
   old_tape_des = tape_des;
   tape_offline (tape_des);
@@ -869,7 +903,7 @@ get_next_reel (int tape_des)
 
 	  str_res = ds_fgets (tty_in, &new_name);
 	  if (str_res == NULL || str_res[0] == '\0')
-	    exit (1);
+	    exit (PAXEXIT_FAILURE);
 	  next_archive_name = str_res;
 
 	  tape_des = open_archive (next_archive_name);
@@ -888,7 +922,7 @@ get_next_reel (int tape_des)
      the archive.  */
 
   if (tape_des != old_tape_des)
-    error (1, 0, _("internal error: tape descriptor changed from %d to %d"),
+    error (PAXEXIT_FAILURE, 0, _("internal error: tape descriptor changed from %d to %d"),
 	   old_tape_des, tape_des);
 
   free (new_name.ds_string);
@@ -957,146 +991,6 @@ umasked_symlink (char *name1, char *name2, int mode)
 }
 #endif /* SYMLINK_USES_UMASK */
 
-#ifdef HPUX_CDF
-/* When we create a cpio archive we mark CDF's by putting an extra `/'
-   after their component name so we can distinguish the CDF's when we
-   extract the archive (in case the "hidden" directory's files appear
-   in the archive before the directory itself).  E.g., in the path
-   "a/b+/c", if b+ is a CDF, we will write this path as "a/b+//c" in
-   the archive so when we extract the archive we will know that b+
-   is actually a CDF, and not an ordinary directory whose name happens
-   to end in `+'.  We also do the same thing internally in copypass.c.  */
-
-
-/* Take an input pathname and check it for CDF's.  Insert an extra
-   `/' in the pathname after each "hidden" directory.  If we add
-   any `/'s, return a malloced string instead of the original input
-   string.
-   FIXME: This creates a memory leak.
-*/
-
-char *
-add_cdf_double_slashes (char *input_name)
-{
-  static char *ret_name = NULL;	/* re-usuable return buffer (malloc'ed)  */
-  static int ret_size = -1;	/* size of return buffer.  */
-  char *p;
-  char *q;
-  int n;
-  struct stat dir_stat;
-
-  /*  Search for a `/' preceeded by a `+'.  */
-
-  for (p = input_name; *p != '\0'; ++p)
-    {
-      if ( (*p == '+') && (*(p + 1) == '/') )
-	break;
-    }
-
-  /* If we didn't find a `/' preceeded by a `+' then there are
-     no CDF's in this pathname.  Return the original pathname.  */
-
-  if (*p == '\0')
-    return input_name;
-
-  /* There was a `/' preceeded by a `+' in the pathname.  If it is a CDF 
-     then we will need to copy the input pathname to our return
-     buffer so we can insert the extra `/'s.  Since we can't tell
-     yet whether or not it is a CDF we will just always copy the
-     string to the return buffer.  First we have to make sure the
-     buffer is large enough to hold the string and any number of
-     extra `/'s we might add.  */
-
-  n = 2 * (strlen (input_name) + 1);
-  if (n >= ret_size)
-    {
-      if (ret_size < 0)
-	ret_name = (char *) malloc (n);
-      else
-	ret_name = (char *)realloc (ret_name, n);
-      ret_size = n;
-    }
-
-  /* Clear the `/' after this component, so we can stat the pathname 
-     up to and including this component.  */
-  ++p;
-  *p = '\0';
-  if ((*xstat) (input_name, &dir_stat) < 0)
-    {
-      stat_error (input_name);
-      return input_name;
-    }
-
-  /* Now put back the `/' after this component and copy the pathname up to
-     and including this component and its trailing `/' to the return
-     buffer.  */
-  *p++ = '/';
-  strncpy (ret_name, input_name, p - input_name);
-  q = ret_name + (p - input_name);
-
-  /* If it was a CDF, add another `/'.  */
-  if (S_ISDIR (dir_stat.st_mode) && (dir_stat.st_mode & 04000) )
-    *q++ = '/';
-
-  /* Go through the rest of the input pathname, copying it to the
-     return buffer, and adding an extra `/' after each CDF.  */
-  while (*p != '\0')
-    {
-      if ( (*p == '+') && (*(p + 1) == '/') )
-	{
-	  *q++ = *p++;
-
-	  *p = '\0';
-	  if ((*xstat) (input_name, &dir_stat) < 0)
-	    {
-	      stat_error (input_name);
-	      return input_name;
-	    }
-	  *p = '/';
-
-	  if (S_ISDIR (dir_stat.st_mode) && (dir_stat.st_mode & 04000) )
-	    *q++ = '/';
-	}
-      *q++ = *p++;
-    }
-  *q = '\0';
-
-  return ret_name;
-}
-
-/* Is the last parent directory (e.g., c in a/b/c/d) a CDF?  If the
-   directory name ends in `+' and is followed by 2 `/'s instead of 1
-   then it is.  This is only the case for cpio archives, but we don't
-   have to worry about tar because tar always has the directory before
-   its files (or else we lose).  */
-int
-islastparentcdf (char *path)
-{
-  char *newpath;
-  char *slash;
-  int slash_count;
-  int length;			/* Length of result, not including NUL.  */
-
-  slash = strrchr (path, '/');
-  if (slash == 0)
-    return 0;
-  else
-    {
-      slash_count = 0;
-      while (slash > path && *slash == '/')
-	{
-	  ++slash_count;
-	  --slash;
-	}
-
-
-      if ( (*slash == '+') && (slash_count >= 2) )
-	return 1;
-    }
-  return 0;
-}
-#endif
-
 #define DISKBLOCKSIZE	(512)
 
 static int
@@ -1111,107 +1005,95 @@ buf_all_zeros (char *buf, int bufsize)
   return 1;
 }
 
-int delayed_seek_count = 0;
+/* Write NBYTE bytes from BUF to file descriptor FILDES, trying to
+   create holes instead of writing blockfuls of zeros.
+   
+   Return the number of bytes written (including bytes in zero
+   regions) on success, -1 on error.
 
-/* Write NBYTE bytes from BUF to remote tape connection FILDES.
-   Return the number of bytes written on success, -1 on error.  */
+   If FLUSH is set, make sure the trailing zero region is flushed
+   on disk.
+*/
 
-static int
-sparse_write (int fildes, char *buf, unsigned int nbyte)
+static ssize_t
+sparse_write (int fildes, char *buf, size_t nbytes, bool flush)
 {
-  int complete_block_count;
-  int leftover_bytes_count;
-  int seek_count;
-  int write_count;
-  char *cur_write_start;
-  int lseek_rc;
-  int write_rc;
-  int i;
-  enum { begin, in_zeros, not_in_zeros } state;
+  size_t nwritten = 0;
+  ssize_t n;
+  char *start_ptr = buf;
 
-  complete_block_count = nbyte / DISKBLOCKSIZE;
-  leftover_bytes_count = nbyte % DISKBLOCKSIZE;
+  static off_t delayed_seek_count = 0;
+  off_t seek_count = 0;
 
-  if (delayed_seek_count != 0)
-    state = in_zeros;
-  else
-    state = begin;
-
-  seek_count = delayed_seek_count;
-
-  for (i = 0; i < complete_block_count; ++i)
+  enum { begin, in_zeros, not_in_zeros } state =
+			   delayed_seek_count ? in_zeros : begin;
+  
+  while (nbytes)
     {
-      switch (state)
+      size_t rest = nbytes;
+
+      if (rest < DISKBLOCKSIZE)
+	/* Force write */
+	state = not_in_zeros;
+      else
 	{
-	  case begin :
-	    if (buf_all_zeros (buf, DISKBLOCKSIZE))
-	      {
-		seek_count = DISKBLOCKSIZE;
-		state = in_zeros;
-	      }
-	    else
-	      {
-		cur_write_start = buf;
-		write_count = DISKBLOCKSIZE;
-		state = not_in_zeros;
-	      }
-	    buf += DISKBLOCKSIZE;
-	    break;
-	    
-	  case in_zeros :
-	    if (buf_all_zeros (buf, DISKBLOCKSIZE))
-	      {
-		seek_count += DISKBLOCKSIZE;
-	      }
-	    else
-	      {
-		lseek (fildes, seek_count, SEEK_CUR);
-		cur_write_start = buf;
-		write_count = DISKBLOCKSIZE;
-		state = not_in_zeros;
-	      }
-	    buf += DISKBLOCKSIZE;
-	    break;
-	    
-	  case not_in_zeros :
-	    if (buf_all_zeros (buf, DISKBLOCKSIZE))
-	      {
-		write_rc = write (fildes, cur_write_start, write_count);
-		seek_count = DISKBLOCKSIZE;
-		state = in_zeros;
-	      }
-	    else
-	      {
-		write_count += DISKBLOCKSIZE;
-	      }
-	    buf += DISKBLOCKSIZE;
-	    break;
+	  if (buf_all_zeros (buf, rest))
+	    {
+	      if (state == not_in_zeros)
+		{
+		  ssize_t bytes = buf - start_ptr + rest;
+		  
+		  n = write (fildes, start_ptr, bytes);
+		  if (n == -1)
+		    return -1;
+		  nwritten += n;
+		  if (n < bytes)
+		    return nwritten + seek_count;
+		  start_ptr = NULL;
+		}
+	      else
+		seek_count += rest;
+	      state = in_zeros;
+	    }
+	  else
+	    {
+	      seek_count += delayed_seek_count;
+	      if (lseek (fildes, seek_count, SEEK_CUR) == -1)
+		return -1;
+	      delayed_seek_count = seek_count = 0;
+	      state = not_in_zeros;
+	      start_ptr = buf;
+	    }
 	}
+      buf += rest;
+      nbytes -= rest;
     }
 
-  switch (state)
+  if (state != in_zeros)
     {
-      case begin :
-      case in_zeros :
-	delayed_seek_count = seek_count;
-	break;
-	
-      case not_in_zeros :
-	write_rc = write (fildes, cur_write_start, write_count);
-	delayed_seek_count = 0;
-	break;
-    }
+      seek_count += delayed_seek_count;
+      if (seek_count && lseek (fildes, seek_count, SEEK_CUR) == -1)
+	return -1;
+      delayed_seek_count = seek_count = 0;
 
-  if (leftover_bytes_count != 0)
-    {
-      if (delayed_seek_count != 0)
-	{
-	  lseek_rc = lseek (fildes, delayed_seek_count, SEEK_CUR);
-	  delayed_seek_count = 0;
-	}
-      write_rc = write (fildes, buf, leftover_bytes_count);
+      n = write (fildes, start_ptr, buf - start_ptr);
+      if (n == -1)
+	return n;
+      nwritten += n;
     }
-  return nbyte;
+  delayed_seek_count += seek_count;
+
+  if (flush && delayed_seek_count)
+    {
+      if (lseek (fildes, delayed_seek_count - 1, SEEK_CUR) == -1)
+	return -1;
+      n = write (fildes, "", 1);
+      if (n != 1)
+	return n;
+      delayed_seek_count = 0;
+    }      
+  
+  return nwritten + seek_count;
 }
 
 #define CPIO_UID(uid) (set_owner_flag ? set_owner : (uid))
@@ -1220,9 +1102,8 @@ sparse_write (int fildes, char *buf, unsigned int nbyte)
 void
 stat_to_cpio (struct cpio_file_stat *hdr, struct stat *st)
 {
-  hdr->c_dev_maj = major (st->st_dev);
-  hdr->c_dev_min = minor (st->st_dev);
-  hdr->c_ino = st->st_ino;
+  get_inode_and_dev (hdr, st);
+
   /* For POSIX systems that don't define the S_IF macros,
      we can't assume that S_ISfoo means the standard Unix
      S_IFfoo bit(s) are set.  So do it manually, with a
@@ -1256,9 +1137,9 @@ stat_to_cpio (struct cpio_file_stat *hdr, struct stat *st)
   else if (S_ISNWK (st->st_mode))
     hdr->c_mode |= CP_IFNWK;
 #endif
+  hdr->c_nlink = st->st_nlink;
   hdr->c_uid = CPIO_UID (st->st_uid);
   hdr->c_gid = CPIO_GID (st->st_gid);
-  hdr->c_nlink = st->st_nlink;
   hdr->c_rdev_maj = major (st->st_rdev);
   hdr->c_rdev_min = minor (st->st_rdev);
   hdr->c_mtime = st->st_mtime;
@@ -1267,6 +1148,49 @@ stat_to_cpio (struct cpio_file_stat *hdr, struct stat *st)
   hdr->c_tar_linkname = NULL;
 }
 
+void
+cpio_to_stat (struct stat *st, struct cpio_file_stat *hdr)
+{
+  memset (st, 0, sizeof (*st));
+  st->st_dev = makedev (hdr->c_dev_maj, hdr->c_dev_min);
+  st->st_ino = hdr->c_ino;
+  st->st_mode = hdr->c_mode & 0777;
+  if (hdr->c_mode & CP_IFREG)
+    st->st_mode |= S_IFREG;
+  else if (hdr->c_mode & CP_IFDIR)
+    st->st_mode |= S_IFDIR;
+#ifdef S_IFBLK
+  else if (hdr->c_mode & CP_IFBLK)
+    st->st_mode |= S_IFBLK;
+#endif
+#ifdef S_IFCHR
+  else if (hdr->c_mode & CP_IFCHR)
+    st->st_mode |= S_IFCHR;
+#endif
+#ifdef S_IFFIFO
+  else if (hdr->c_mode & CP_IFIFO)
+    st->st_mode |= S_IFIFO;
+#endif
+#ifdef S_IFLNK
+  else if (hdr->c_mode & CP_IFLNK)
+    st->st_mode |= S_IFLNK;
+#endif
+#ifdef S_IFSOCK
+  else if (hdr->c_mode & CP_IFSOCK)
+    st->st_mode |= S_IFSOCK;
+#endif
+#ifdef S_IFNWK
+  else if (hdr->c_mode & CP_IFNWK)
+    st->st_mode |= S_IFNWK;
+#endif
+  st->st_nlink = hdr->c_nlink;
+  st->st_uid = CPIO_UID (hdr->c_uid);
+  st->st_gid = CPIO_GID (hdr->c_gid);
+  st->st_rdev = makedev (hdr->c_rdev_maj, hdr->c_rdev_min);
+  st->st_mtime = hdr->c_mtime;
+  st->st_size = hdr->c_filesize;
+}
+
 #ifndef HAVE_FCHOWN
 # define HAVE_FCHOWN 0
 #endif
@@ -1289,7 +1213,7 @@ fchmod_or_chmod (int fd, const char *name, mode_t mode)
   if (HAVE_FCHMOD && fd != -1)
     return fchmod (fd, mode);
   else
-    return chmod(name, mode);
+    return chmod (name, mode);
 }
 
 void
@@ -1306,12 +1230,6 @@ set_perms (int fd, struct cpio_file_stat *header)
   /* chown may have turned off some permissions we wanted. */
   if (fchmod_or_chmod (fd, header->c_name, header->c_mode) < 0)
     chmod_error_details (header->c_name, header->c_mode);
-#ifdef HPUX_CDF
-  if ((header->c_mode & CP_IFMT) && cdf_flag)
-    /* Once we "hide" the directory with the chmod(),
-       we have to refer to it using name+ instead of name.  */
-    file_hdr->c_name [cdf_char] = '+';
-#endif
   if (retain_time_flag)
     set_file_times (fd, header->c_name, header->c_mtime, header->c_mtime);
 }
@@ -1329,12 +1247,32 @@ set_file_times (int fd,
 
   /* Silently ignore EROFS because reading the file won't have upset its 
      timestamp if it's on a read-only filesystem. */
-  if (gl_futimens (fd, name, ts) < 0 && errno != EROFS)
+  if (fdutimens (fd, name, ts) < 0 && errno != EROFS)
     utime_error (name);
 }
 
+/* Reallocate file_hdr->c_name to accomodate len bytes (including final \0) */
+void
+cpio_realloc_c_name (struct cpio_file_stat *file_hdr, size_t len)
+{
+  while (file_hdr->c_name_buflen < len)
+    file_hdr->c_name = x2realloc (file_hdr->c_name, &file_hdr->c_name_buflen);
+}
+
+void
+cpio_set_c_name (struct cpio_file_stat *file_hdr, char *name)
+{
+  size_t len = strlen (name) + 1;
+
+  cpio_realloc_c_name (file_hdr, len);
+  file_hdr->c_namesize = len;
+  memmove (file_hdr->c_name, name, len);
+}
+
 /* Do we have to ignore absolute paths, and if so, does the filename
-   have an absolute path?  */
+   have an absolute path?  Before calling this function make sure that the
+   allocated NAME buffer has capacity at least 2 bytes. */
+
 void
 cpio_safer_name_suffix (char *name, bool link_target, bool absolute_names,
 			bool strip_leading_dots)
@@ -1349,6 +1287,10 @@ cpio_safer_name_suffix (char *name, bool link_target, bool absolute_names,
 	  ++p;
       }
   if (p != name)
+    /* The 'p' string is shortened version of 'name' with one exception;  when
+       the 'name' points to an empty string (buffer where name[0] == '\0') the
+       'p' then points to static string ".".  So caller needs to ensure there
+       are at least two bytes available in 'name' buffer so memmove succeeds. */
     memmove (name, p, (size_t)(strlen (p) + 1));
 }
 
@@ -1375,28 +1317,37 @@ struct delayed_set_stat
 static struct delayed_set_stat *delayed_set_stat_head;
 
 void
-delay_set_stat (char const *file_name, struct stat *st,
-		mode_t invert_permissions)
+delay_cpio_set_stat (struct cpio_file_stat *file_stat,
+		     mode_t invert_permissions)
 {
-  size_t file_name_len = strlen (file_name);
+  size_t file_name_len = strlen (file_stat->c_name);
   struct delayed_set_stat *data =
     xmalloc (sizeof (struct delayed_set_stat) + file_name_len + 1);
   data->next = delayed_set_stat_head;
-  memset (&data->stat, 0, sizeof data->stat);
-  stat_to_cpio (&data->stat, st);
+  memcpy (&data->stat, file_stat, sizeof data->stat);
   data->stat.c_name = (char*) (data + 1);
-  strcpy (data->stat.c_name, file_name);
+  strcpy (data->stat.c_name, file_stat->c_name);
   data->invert_permissions = invert_permissions;
   delayed_set_stat_head = data;
 }
 
+void
+delay_set_stat (char const *file_name, struct stat *st,
+		mode_t invert_permissions)
+{
+  struct cpio_file_stat fs;
+
+  stat_to_cpio (&fs, st);
+  fs.c_name = (char*) file_name;
+  delay_cpio_set_stat (&fs, invert_permissions);
+}
+
 /* Update the delayed_set_stat info for an intermediate directory
    created within the file name of DIR.  The intermediate directory turned
    out to be the same as this directory, e.g. due to ".." or symbolic
    links.  *DIR_STAT_INFO is the status of the directory.  */
-void
-repair_delayed_set_stat (char const *dir,
-			 struct stat *dir_stat_info)
+int
+repair_inter_delayed_set_stat (struct stat *dir_stat_info)
 {
   struct delayed_set_stat *data;
   for (data = delayed_set_stat_head; data; data = data->next)
@@ -1405,7 +1356,7 @@ repair_delayed_set_stat (char const *dir,
       if (stat (data->stat.c_name, &st) != 0)
 	{
 	  stat_error (data->stat.c_name);
-	  return;
+	  return -1;
 	}
 
       if (st.st_dev == dir_stat_info->st_dev
@@ -1415,12 +1366,31 @@ repair_delayed_set_stat (char const *dir,
 	  data->invert_permissions =
 	    ((dir_stat_info->st_mode ^ st.st_mode)
 	     & MODE_RWX & ~ newdir_umask);
-	  return;
+	  return 0;
 	}
     }
+  return 1;
+}
 
-  ERROR ((0, 0, _("%s: Unexpected inconsistency when making directory"),
-	  quotearg_colon (dir)));
+/* Update the delayed_set_stat info for a directory matching
+   FILE_HDR.
+
+   Return 0 if such info was found, 1 otherwise. */
+int
+repair_delayed_set_stat (struct cpio_file_stat *file_hdr)
+{
+  struct delayed_set_stat *data;
+  for (data = delayed_set_stat_head; data; data = data->next)
+    {
+      if (strcmp (file_hdr->c_name, data->stat.c_name) == 0)
+	{
+	  data->invert_permissions = 0;
+	  memcpy (&data->stat, file_hdr,
+		  offsetof (struct cpio_file_stat, c_name));
+	  return 0;
+	}
+    }
+  return 1;
 }
 
 void
@@ -1439,4 +1409,137 @@ apply_delayed_set_stat ()
     }
 }
 
+
+static int
+cpio_mkdir (struct cpio_file_stat *file_hdr, int *setstat_delayed)
+{
+  int rc;
+  mode_t mode = file_hdr->c_mode;
+  
+  if (!(file_hdr->c_mode & S_IWUSR))
+    {
+      rc = mkdir (file_hdr->c_name, mode | S_IWUSR);
+      if (rc == 0)
+	{
+	  delay_cpio_set_stat (file_hdr, 0);
+	  *setstat_delayed = 1;
+	}
+    }
+  else
+    {
+      rc = mkdir (file_hdr->c_name, mode);
+      *setstat_delayed = 0;
+    }
+  return rc;
+}
 
+int
+cpio_create_dir (struct cpio_file_stat *file_hdr, int existing_dir)
+{
+  int res;			/* Result of various function calls.  */
+  int setstat_delayed = 0;
+  
+  if (to_stdout_option)
+    return 0;
+  
+  /* Strip any trailing `/'s off the filename; tar puts
+     them on.  We might as well do it here in case anybody
+     else does too, since they cause strange things to happen.  */
+  strip_trailing_slashes (file_hdr->c_name);
+
+  /* Ignore the current directory.  It must already exist,
+     and we don't want to change its permission, ownership
+     or time.  */
+  if (file_hdr->c_name[0] == '.' && file_hdr->c_name[1] == '\0')
+    {
+      return 0;
+    }
+
+  if (!existing_dir)
+    res = cpio_mkdir (file_hdr, &setstat_delayed);
+  else
+    res = 0;
+  if (res < 0 && create_dir_flag)
+    {
+      create_all_directories (file_hdr->c_name);
+      res = cpio_mkdir (file_hdr, &setstat_delayed);
+    }
+  if (res < 0)
+    {
+      /* In some odd cases where the file_hdr->c_name includes `.',
+	 the directory may have actually been created by
+	 create_all_directories(), so the mkdir will fail
+	 because the directory exists.  If that's the case,
+	 don't complain about it.  */
+      struct stat file_stat;
+      if (errno != EEXIST)
+	{
+	  mkdir_error (file_hdr->c_name);
+	  return -1;
+	}
+      if (lstat (file_hdr->c_name, &file_stat))
+	{
+	  stat_error (file_hdr->c_name);
+	  return -1;
+	}
+      if (!(S_ISDIR (file_stat.st_mode)))
+	{
+	  error (0, 0, _("%s is not a directory"),
+		 quotearg_colon (file_hdr->c_name));
+	  return -1;
+	}
+    }
+
+  if (!setstat_delayed && repair_delayed_set_stat (file_hdr))
+    set_perms (-1, file_hdr);
+  return 0;
+}
+
+void
+change_dir ()
+{
+  if (change_directory_option && chdir (change_directory_option))
+    {
+      if (errno == ENOENT && create_dir_flag)
+	{
+	  if (make_path (change_directory_option, -1, -1,
+			 (warn_option & CPIO_WARN_INTERDIR) ?
+			 _("Creating directory `%s'") : NULL))
+	    exit (PAXEXIT_FAILURE);
+
+	  if (chdir (change_directory_option) == 0)
+	    return;
+	}
+      error (PAXEXIT_FAILURE, errno,
+	     _("cannot change to directory `%s'"), change_directory_option);
+    }
+}
+
+/* Return true if the archive format ARF stores inode numbers */
+int
+arf_stores_inode_p (enum archive_format arf)
+{
+  switch (arf)
+    {
+    case arf_tar:
+    case arf_ustar:
+      return 0;
+
+    default:
+      break;
+    }
+  return 1;
+}
+  
+void
+cpio_file_stat_init (struct cpio_file_stat *file_hdr)
+{
+  memset (file_hdr, 0, sizeof (*file_hdr));
+}
+
+void
+cpio_file_stat_free (struct cpio_file_stat *file_hdr)
+{
+  free (file_hdr->c_name);
+  cpio_file_stat_init (file_hdr);
+}
